"""Defines functions to plot positioned ZX graphs and correlation surfaces on 3D axes."""

from dataclasses import astuple

import matplotlib.pyplot as plt
import numpy
import numpy.typing as npt
from matplotlib.figure import Figure
from mpl_toolkits.mplot3d.axes3d import Axes3D
from pyzx import draw_3d
from pyzx.graph.graph_s import GraphS
from pyzx.pauliweb import PauliWeb

from tqec.computation.correlation import CorrelationSurface
from tqec.interop.color import RGBA, TQECColor
from tqec.interop.pyzx.positioned import PositionedZX
from tqec.interop.pyzx.utils import is_boundary, is_hardmard, is_s, is_z_no_phase
from tqec.utils.position import Position3D


def _node_color(g: GraphS, v: int) -> RGBA:  # pragma: no cover
    assert not is_boundary(g, v)
    if is_s(g, v):
        return TQECColor.Y.rgba
    if is_z_no_phase(g, v):
        return TQECColor.Z.rgba
    return TQECColor.X.rgba


def _positions_array(*positions: Position3D) -> npt.NDArray[numpy.int_]:  # pragma: no cover
    return numpy.array([astuple(p) for p in positions]).T


def draw_positioned_zx_graph_on(
    graph: PositionedZX,
    ax: Axes3D,
    *,
    node_size: int = 400,
    hadamard_size: int = 200,
    edge_width: int = 1,
) -> None:
    """Draw the :py:class:`~tqec.interop.pyzx.PositionedZX` on the provided axes.

    Args:
        graph: The positioned ZX graph to draw.
        ax: a 3-dimensional ax to draw on.
        node_size: The size of the node. Default is 400.
        hadamard_size: The size of the Hadamard transition. Default is 200.
        edge_width: The width of the edge. Default is 1.

    """
    g = graph.g
    pmap = graph.positions
    vis_nodes = [n for n in g.vertices() if not is_boundary(g, n)]
    vis_nodes_array = _positions_array(*[pmap[n] for n in vis_nodes])
    if vis_nodes_array.size > 0:
        ax.scatter(
            *vis_nodes_array,
            s=node_size,
            c=[_node_color(g, n).as_floats() for n in vis_nodes],
            alpha=1.0,
            edgecolors="black",
        )

    for edge in g.edges():
        pos_array = _positions_array(pmap[edge[0]], pmap[edge[1]])
        ax.plot(
            *pos_array,
            color="tab:gray",
            linewidth=edge_width,
        )
        if is_hardmard(g, edge):
            hadamard_position = numpy.mean(pos_array, axis=1)
            # use yellow square to indicate Hadamard transition
            ax.scatter(
                *hadamard_position,
                s=hadamard_size,
                c="yellow",
                alpha=1.0,
                edgecolors="black",
                marker="s",
            )
    ax.grid(False)
    for dim in (ax.xaxis, ax.yaxis, ax.zaxis):
        dim.set_ticks([])
    x_limits, y_limits, z_limits = ax.get_xlim3d(), ax.get_ylim3d(), ax.get_zlim3d()  # type: ignore

    plot_radius = 0.5 * max(abs(limits[1] - limits[0]) for limits in [x_limits, y_limits, z_limits])

    ax.set_xlim3d([numpy.mean(x_limits) - plot_radius, numpy.mean(x_limits) + plot_radius])
    ax.set_ylim3d([numpy.mean(y_limits) - plot_radius, numpy.mean(y_limits) + plot_radius])
    ax.set_zlim3d([numpy.mean(z_limits) - plot_radius, numpy.mean(z_limits) + plot_radius])


def draw_correlation_surface_on(
    correlation_surface: CorrelationSurface,
    graph: PositionedZX,
    ax: Axes3D,
    correlation_edge_width: int = 3,
) -> None:
    """Draw the correlation surface on the provided axes.

    Args:
        correlation_surface: The correlation surface to draw.
        graph: The positioned ZX graph to draw the correlation surface on.
        ax: The 3-dimensional ax to draw on.
        correlation_edge_width: The width of the correlation edges. Default is 3.

    """
    if correlation_surface.is_single_node:
        return

    pmap = graph.positions
    pauli_web = correlation_surface.to_pauli_web(graph.g)

    for edge, pauli in pauli_web.half_edges().items():
        up, vp = pmap[edge[0]], pmap[edge[1]]
        pos_array = _positions_array(up, vp)
        middle = numpy.mean(pos_array, axis=1).reshape(3, 1)
        start = _positions_array(up)
        ax.plot(
            *numpy.hstack([start, middle]),
            color=TQECColor(pauli).rgba.as_floats(),
            linewidth=correlation_edge_width,
        )


def plot_positioned_zx_graph(
    graph: PositionedZX,
    *,
    figsize: tuple[float, float] = (5, 6),
    title: str | None = None,
    node_size: int = 400,
    hadamard_size: int = 200,
    edge_width: int = 1,
) -> tuple[Figure, Axes3D]:
    """Plot the :py:class:`~tqec.interop.pyzx.positioned.PositionedZX` using matplotlib.

    Args:
        graph: The positioned ZX graph to plot.
        figsize: The figure size. Default is (5, 6).
        title: The title of the plot. Default to the name of the graph.
        node_size: The size of the node in the plot. Default is 400.
        hadamard_size: The size of the Hadamard square in the plot. Default is 200.
        edge_width: The width of the edge in the plot. Default is 1.

    Returns:
        A tuple of the figure and the axes.

    """
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111, projection="3d")

    draw_positioned_zx_graph_on(
        graph,
        ax,
        node_size=node_size,
        hadamard_size=hadamard_size,
        edge_width=edge_width,
    )

    if title:
        ax.set_title(title)
    fig.tight_layout()
    return fig, ax


def pyzx_draw_positioned_zx_3d(
    g: PositionedZX,
    id_labels: bool = True,
    pauli_web: PauliWeb[int, tuple[int, int]] | None = None,
) -> None:
    """Draw the positioned ZX graph in 3D with ``pyzx.draw_3d``.

    Args:
        g: The positioned ZX graph to draw.
        id_labels: Whether to show the vertex id labels. Default is True.
        pauli_web: The Pauli web to draw. Default is None.

    """
    plot_g = g.g.clone()
    for v in plot_g.vertices():
        position = g.positions[v]
        plot_g.set_qubit(v, position.x)
        plot_g.set_row(v, position.y)
        plot_g.set_vdata(v, "z", position.z)
    draw_3d(plot_g, labels=id_labels, pauli_web=pauli_web)



from __future__ import annotations

from enum import Enum, auto


class Orientation(Enum):
    """Either horizontal or vertical orientation."""

    HORIZONTAL = auto()
    VERTICAL = auto()

    def flip(self) -> Orientation:
        """Flip ``self``, returning the opposite orientation."""
        return Orientation.HORIZONTAL if self == Orientation.VERTICAL else Orientation.VERTICAL


class Basis(Enum):
    X = "X"
    Z = "Z"

    def flipped(self) -> Basis:
        """Flip ``self``, returning the opposite basis."""
        return Basis.X if self == Basis.Z else Basis.Z

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.{self.value}"

    def __lt__(self, other: Basis) -> bool:
        return self.value < other.value


class PatchStyle(Enum):
    FixedBulk = auto()
    FixedBoundaryParity = auto()



"""Utility functions for PyZX interop."""

from fractions import Fraction

from pyzx.graph.graph_s import GraphS
from pyzx.utils import EdgeType, FractionLike, VertexType, vertex_is_zx

from tqec.computation.cube import CubeKind, Port, ZXCube
from tqec.utils.enums import Basis


def is_zx_no_phase(g: GraphS, v: int) -> bool:
    """Check if a vertex in a PyZX graph is a Z/X spider with phase 0."""
    return vertex_is_zx(g.type(v)) and g.phase(v) == 0


def is_z_no_phase(g: GraphS, v: int) -> bool:
    """Check if a vertex in a PyZX graph is a Z spider with phase 0."""
    return g.type(v) is VertexType.Z and g.phase(v) == 0


def is_x_no_phase(g: GraphS, v: int) -> bool:
    """Check if a vertex in a PyZX graph is a X spider with phase 0."""
    return g.type(v) is VertexType.X and g.phase(v) == 0


def is_boundary(g: GraphS, v: int) -> bool:
    """Check if a vertex in a PyZX graph is a boundary type spider."""
    return g.type(v) is VertexType.BOUNDARY


def is_s(g: GraphS, v: int) -> bool:
    """Check if a vertex in a PyZX graph is a S node."""
    return g.type(v) is VertexType.Z and g.phase(v) == Fraction(1, 2)


def is_hardmard(g: GraphS, edge: tuple[int, int]) -> bool:
    """Check if an edge in a PyZX graph is a Hadamard edge."""
    return g.edge_type(edge) is EdgeType.HADAMARD


def cube_kind_to_zx(kind: CubeKind) -> tuple[VertexType, FractionLike]:
    """Convert the cube kind to the corresponding PyZX vertex type and phase.

    The conversion is as follows:

    - Port -> BOUNDARY spider with phase 0.
    - YHalfCube -> Z spider with phase 1/2.
    - ZXCube -> Z spider with phase 0 if it has only one Z basis boundary,
        otherwise X spider with phase 0.

    Args:
        kind: The cube kind to be converted.

    Returns:
        A tuple of vertex type and spider phase.

    """
    if isinstance(kind, ZXCube):
        if sum(basis == Basis.Z for basis in kind.as_tuple()) == 1:
            return VertexType.Z, 0
        return VertexType.X, 0
    if isinstance(kind, Port):
        return VertexType.BOUNDARY, 0
    else:  # isinstance(kind, YHalfCube)
        return VertexType.Z, Fraction(1, 2)



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass, field
from enum import Enum
from typing import Any

from tqec.circuit.schedule.schedule import Schedule
from tqec.utils.enums import Basis


class PauliBasis(Enum):
    X = "x"
    Y = "y"
    Z = "z"

    def __str__(self) -> str:
        return self.value  # pragma: no cover

    def to_extended_basis(self) -> ExtendedBasis:
        """Return ``self`` as an extended basis."""
        return ExtendedBasis(self.value)


class ExtendedBasis(Enum):
    X = PauliBasis.X.value
    Y = PauliBasis.Y.value
    Z = PauliBasis.Z.value
    H = "h"

    def __str__(self) -> str:
        return self.value  # pragma: no cover


@dataclass(frozen=True)
class RPNG:
    """Represent a single ``RPNG`` string.

    ## Format specification

    The ``RPNG`` string is a standard format used in ``tqec`` to unambiguously
    describe the action(s) being performed on a single data qubit. It is a
    4-character string. See each attribute docstring for more details on the
    possible values for each character.

    ## Example

    The following ::

        -z1-
        rpng

    represents a data-qubit with a ``CZ`` gate applied at timestep ``1``.

    Attributes:
        r: reset basis (``x``, ``y`` or ``z``), ``h`` or ``-``.
        p: controlled operation target basis (``x`` means ``CNOT`` controlled
            on the ancilla and targeting the data qubit, ``y`` means ``CY``,
            ``z`` means ``CZ``).
        n: time step at which the 2-qubit operation described by ``p`` should
            be applied. Should be a 1-digit positive integer, typically in
            ``[1, 5]``.
        g: measure basis (``x``, ``y`` or ``z``), ``h`` or ``-``.

    """

    r: ExtendedBasis | None
    p: PauliBasis | None
    n: int | None
    g: ExtendedBasis | None

    @classmethod
    def from_string(cls, rpng_string: str) -> RPNG:
        """Initialize the RPNG object from a 4-character string.

        Raises:
            ValueError: if an invalid ``rpng_string`` is provided.

        """
        if len(rpng_string) != 4:
            raise ValueError("The rpng string must be exactly 4-character long.")
        r_str, p_str, n_str, g_str = tuple(rpng_string)
        # Convert the characters into the enum attributes (or raise error).
        r = ExtendedBasis(r_str) if r_str in ExtendedBasis._value2member_map_ else None
        p = PauliBasis(p_str) if p_str in PauliBasis._value2member_map_ else None
        n = int(n_str) if n_str.isdigit() else None
        g = ExtendedBasis(g_str) if g_str in ExtendedBasis._value2member_map_ else None
        # Raise error if anythiong but '-' was used to indicate None.
        if not r and r_str != "-":
            raise ValueError("Unacceptable character for the R field.")
        if not p and p_str != "-":
            raise ValueError("Unacceptable character for the P field.")
        if not n and n_str != "-":
            raise ValueError("Unacceptable character for the N field.")
        if not g and g_str != "-":
            raise ValueError("Unacceptable character for the G field.")
        return cls(r, p, n, g)

    def get_r_op(self) -> str | None:
        """Get the reset operation or Hadamard."""
        op = self.r
        if op is None:
            return None
        elif op.value in PauliBasis._value2member_map_:
            return f"R{op.value.upper()}"
        else:
            return f"{op.value.upper()}"

    def get_g_op(self) -> str | None:
        """Get the measurement operation or Hadamard."""
        op = self.g
        if op is None:
            return None
        elif op.value in PauliBasis._value2member_map_:
            return f"M{op.value.upper()}"
        else:
            return f"{op.value.upper()}"

    @property
    def is_null(self) -> bool:
        """Check if the RPNG object is null, i.e. all fields are None."""
        return str(self) == "----"

    def __str__(self) -> str:
        r = "-" if self.r is None else self.r.value
        p = "-" if self.p is None else self.p.value
        n = "-" if self.n is None else self.n
        g = "-" if self.g is None else self.g.value
        return f"{r}{p}{n}{g}"


@dataclass(frozen=True)
class RG:
    """Reduced format to represent syndrome qubit operations.

    The ``RG`` format is simply the ``RPNG`` format where ``P`` is
    unconditionally ``-`` and ``N`` is unset.

    Attributes:
        r: reset basis (``x``, ``y`` or ``z``), ``h`` or ``-``.
        g: measure basis (``x``, ``y`` or ``z``), ``h`` or ``-``.

    """

    r: PauliBasis | None
    g: PauliBasis | None

    @classmethod
    def from_string(cls, rg_string: str) -> RG:
        """Initialize the ``RG`` object from a 2-character string."""
        if len(rg_string) != 2:
            raise ValueError("The RG string must be exactly 2-character long.")
        r_str, g_str = tuple(rg_string)

        try:
            r = None if r_str == "-" else PauliBasis(r_str)
            g = None if g_str == "-" else PauliBasis(g_str)
            return cls(r, g)
        except ValueError as err:
            raise ValueError(f"Invalid RG string: '{rg_string}'.") from err

    def __str__(self) -> str:
        return f"{'-' if self.r is None else self.r.value}{'-' if self.g is None else self.g.value}"


@dataclass
class RPNGDescription:
    """Organize the description of a plaquette in RPNG format.

    The corners of the square plaquette are listed following the order:
    top-left, top-right, bottom-left, bottom-right.
    This forms a Z-shaped path on the plaquette corners:

    .. code-block:: text

        +------------+
        | 1 -----> 2 |
        |        /   |
        |      /     |
        |    ∟       |
        | 3 -----> 4 |
        +------------+

    If the ancilla RG description is not specified, it is assumed 'xx'

    Attributes:
        corners: one ``RPNG`` description for each of the four corners of the
            plaquette.
        ancilla: ``RG`` description of the syndrome qubit.

    """

    corners: tuple[RPNG, RPNG, RPNG, RPNG]
    ancilla: RG = field(default=RG(PauliBasis.X, PauliBasis.X))

    def __post_init__(self) -> None:
        """Validate the initialization arguments.

        Constraints:
        - the n values for the corners must be unique
        - the n values for the corners must be larger than 0

        """
        times = []
        for rpng in self.corners:
            if rpng.n:
                if rpng.n < 1:
                    raise ValueError("All n values must be larger than 0.")
                times.append(rpng.n)
        if len(times) != len(set(times)):
            raise ValueError("The n values for the corners must be unique.")

    @classmethod
    def from_string(cls, corners_rpng_string: str) -> RPNGDescription:
        """Initialize the RPNGDescription object from a (16+3)-character string."""
        rpng_objs = tuple([RPNG.from_string(s) for s in corners_rpng_string.split(" ")])
        if len(rpng_objs) != 4:
            raise ValueError("There must be 4 corners in the RPNG description.")
        return cls(rpng_objs)

    @classmethod
    def from_extended_string(cls, ancilla_and_corners_rpng_string: str) -> RPNGDescription:
        """Initialize the RPNGDescription object from a (16+3)-character string."""
        values = ancilla_and_corners_rpng_string.split(" ")
        ancilla_rg = RG.from_string(values[0])
        rpng_objs = tuple([RPNG.from_string(s) for s in values[1:]])
        if len(rpng_objs) != 4:
            raise ValueError("There must be 4 corners in the RPNG description.")
        return cls(rpng_objs, ancilla_rg)

    @classmethod
    def from_basis_and_schedule(
        cls,
        basis: Basis,
        schedule: Sequence[int] | Schedule,
        reset: PauliBasis | None = None,
        measurement: ExtendedBasis | None = None,
    ) -> RPNGDescription:
        """Initialize the RPNGDescription object from a basis and a schedule."""
        r = "-" if reset is None else reset.value
        m = "-" if measurement is None else measurement.value
        rpng_objs = tuple([RPNG.from_string(f"{r}{basis.value.lower()}{s}{m}") for s in schedule])
        if len(rpng_objs) != 4:
            raise ValueError("There must be 4 corners in the RPNG description.")
        return cls(rpng_objs)

    @staticmethod
    def empty() -> RPNGDescription:
        """Return a description of the empty plaquette."""
        return RPNGDescription.from_extended_string("-- ---- ---- ---- ----")

    def get_r_op(self, data_idx: int) -> str | None:
        """Get the reset operation or Hadamard for the specific data qubit."""
        return self.corners[data_idx].get_r_op()

    def get_n(self, data_idx: int) -> int | None:
        """Get the time of the 2Q gate involving the specific data qubit."""
        return self.corners[data_idx].n

    def get_g_op(self, data_idx: int) -> str | None:
        """Get the measurement operation or Hadamard for the specific data qubit."""
        return self.corners[data_idx].get_g_op()

    @property
    def has_reset(self) -> bool:
        """Return ``True`` if ``self`` contains at least one corner with a reset."""
        return any(
            corner.get_r_op()
            not in {
                None,
                "H",
            }
            for corner in self.corners
        )

    @property
    def has_measurement(self) -> bool:
        """Return ``True`` if ``self`` contains at least one corner with a measurement."""
        return any(
            corner.get_g_op()
            not in {
                None,
                "H",
            }
            for corner in self.corners
        )

    def __str__(self) -> str:
        return " ".join(str(rpng) for rpng in self.corners)

    def to_dict(self) -> dict[str, Any]:
        """Return a dictionary representation of the RPNG description.

        The dictionary is intended to be used as a JSON object.

        """
        return {
            "corners": [str(rpng) for rpng in self.corners],
            "ancilla": str(self.ancilla),
        }

    @staticmethod
    def from_dict(data: dict[str, Any]) -> RPNGDescription:
        """Return a RPNGDescription object from its dictionary representation.

        Args:
            data: dictionary with the keys ``corners`` and ``ancilla``.

        Returns:
            a new instance of :class:`RPNGDescription` with the provided
            ``corners`` and ``ancilla``.

        """
        assert len(data["corners"]) == 4, "There must be 4 corners in the RPNG description."
        corners = data["corners"]
        corners = (
            RPNG.from_string(corners[0]),
            RPNG.from_string(corners[1]),
            RPNG.from_string(corners[2]),
            RPNG.from_string(corners[3]),
        )
        ancilla = RG.from_string(data["ancilla"])
        return RPNGDescription(corners, ancilla)



"""Provides interoperability between ``tqec`` and external frameworks / formats."""

from tqec.interop.collada import (
    display_collada_model as display_collada_model,
)
from tqec.interop.collada import (
    read_block_graph_from_dae_file as read_block_graph_from_dae_file,
)
from tqec.interop.collada import (
    write_block_graph_to_dae_file as write_block_graph_to_dae_file,
)
from tqec.interop.color import RGBA as RGBA
from tqec.interop.color import TQECColor as TQECColor
from tqec.interop.pyzx import (
    PositionedZX as PositionedZX,
)
from tqec.interop.pyzx import (
    SynthesisStrategy as SynthesisStrategy,
)
from tqec.interop.pyzx import (
    block_synthesis as block_synthesis,
)
from tqec.interop.pyzx import (
    positioned_block_synthesis as positioned_block_synthesis,
)



from typing_extensions import override

from tqec.compile.blocks.layers.atomic.layout import LayoutLayer
from tqec.compile.tree.node import LayerNode, NodeWalker


class AnnotateCircuitOnLayerNode(NodeWalker):
    def __init__(self, k: int):
        """Walker annotating ``stim.Circuit`` instances implementing the node on each leaf node.

        Args:
            k: scaling factor.

        """
        self._k = k

    @override
    def visit_node(self, node: LayerNode) -> None:
        if not node.is_leaf:
            return
        assert isinstance(node._layer, LayoutLayer)
        node.set_circuit_annotation(self._k, node._layer.to_circuit(self._k))



"""Defines :class:`CorrelationSurface` and functions to find and build them from a ZX graph."""

from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from typing import TYPE_CHECKING

from tqec.computation.block_graph import BlockGraph
from tqec.utils.enums import Basis
from tqec.utils.position import Position3D

if TYPE_CHECKING:
    from pyzx.graph.graph_s import GraphS
    from pyzx.pauliweb import PauliWeb


@dataclass(frozen=True, order=True)
class ZXNode:
    """Represent a node in the ZX graph spanned by the correlation surface.

    Correlation surface is represented by a set of edges in the ZX graph. Each edge
    consists of two nodes. The node id refers to the vertex index in the graph
    and the basis represents the Pauli operator on the half edge incident to the node.

    Attributes:
        id: The index of the vertex in the graph.
        basis: The Pauli operator on the half edge incident to the node.

    """

    id: int
    basis: Basis


@dataclass(frozen=True, order=True)
class ZXEdge:
    """Represent an edge in the ZX graph spanned by the correlation surface.

    Correlation surface is represented by a set of edges in the ZX graph. Each edge
    consists of two nodes. The edge can be decomposed into two half edges, each
    incident to a node. The half edges are labeled with the Pauli operator.
    If the edge has no hadamard effect, the Pauli operators on the two half edges
    should be the same. Otherwise, the Pauli operators should be flipped.

    Note that the Pauli operators are represented by `Basis` enum, which only
    contains `X` and `Z` operators. The `Y` operator is represented by two
    edges with `X` and `Z` operators respectively.

    """

    u: ZXNode
    v: ZXNode

    def __post_init__(self) -> None:
        if self.u.id > self.v.id:
            u, v = self.v, self.u
            object.__setattr__(self, "u", u)
            object.__setattr__(self, "v", v)

    def __iter__(self) -> Iterator[ZXNode]:
        yield self.u
        yield self.v

    def is_self_loop(self) -> bool:
        """Whether the edge is a self-loop edge.

        By definition, a self-loop edge represents a correlation surface within a single node. This
        is an edge case where the ZX graph only contains a single node.

        """
        return self.u.id == self.v.id

    @property
    def has_hadamard(self) -> bool:
        """Whether the edge has a hadamard effect."""
        return self.u.basis != self.v.basis


@dataclass(frozen=True)
class CorrelationSurface:
    """Represent a set of measurements whose values determine the parity of the logical operators.

    A correlation surface in a computation is a set of measurements whose values determine the
    parity of the logical operators at the inputs and outputs associated with the surface.

    Note:
        We use the term "correlation surface", "pauli web" and "observable" interchangeably in
        the library.

    Here we represent the correlation surface by the set of edges it spans in the ZX graph.
    The insight is that the spiders pose parity constraints on the operators supported on
    the incident edges. The flow of the logical operators through the ZX graph, respecting
    the parity constraints, forms the correlation between the inputs and outputs. However,
    the sign of measurement outcomes is neglected in this representation. And we need to
    recover the measurements when instantiating an explicit logical observable from the
    correlation surface.

    Each edge establishes a correlation between the logical operators at the two ends of
    the edge. For example, an edge connecting two Z nodes represents the correlation
    between the logical Z operators at the two nodes.

    Attributes:
        span: A set of ``ZXEdge`` representing the span of the correlation surface.

    """

    span: frozenset[ZXEdge]

    def bases_at(self, v: int) -> set[Basis]:
        """Get the bases of the surfaces present at the vertex."""
        edges = self.edges_at(v)
        bases = set()
        for edge in edges:
            if edge.u.id == v:
                bases.add(edge.u.basis)
            else:
                bases.add(edge.v.basis)
        return bases

    def to_pauli_web(self, g: GraphS) -> PauliWeb[int, tuple[int, int]]:
        """Convert the correlation surface to a Pauli web.

        Args:
            g: The ZX graph the correlation surface is based on.

        Returns:
            A `PauliWeb` representation of the correlation surface.

        """
        # Avoid pulling pyzx when importing that module.
        from tqec.interop.pyzx.correlation import correlation_surface_to_pauli_web  # noqa: PLC0415

        return correlation_surface_to_pauli_web(self, g)

    @staticmethod
    def from_pauli_web(pauli_web: PauliWeb[int, tuple[int, int]]) -> CorrelationSurface:
        """Create a correlation surface from a Pauli web."""
        # Avoid pulling pyzx when importing that module.
        from tqec.interop.pyzx.correlation import pauli_web_to_correlation_surface  # noqa: PLC0415

        return pauli_web_to_correlation_surface(pauli_web)

    @property
    def is_single_node(self) -> bool:
        """Whether the correlation surface contains only a single node.

        This is an edge case where the ZX graph only contains a single node. The span of the
        correlation surface is a self-loop edge at the node.

        """
        return len(self.span) == 1 and next(iter(self.span)).is_self_loop()

    def span_vertices(self) -> set[int]:
        """Return the set of vertices in the correlation surface."""
        return {v.id for edge in self.span for v in edge}

    def edges_at(self, v: int) -> set[ZXEdge]:
        """Return the set of edges incident to the vertex in the correlation surface."""
        return {edge for edge in self.span if any(n.id == v for n in edge)}

    def external_stabilizer(self, io_ports: list[int]) -> str:
        """Get the Pauli operator supported on the given input/output ports.

        Args:
            io_ports: The list of input/output ports to consider.

        Returns:
            The Pauli operator supported on the given ports.

        """
        # Avoid pulling pyzx when importing that module.
        from pyzx.pauliweb import multiply_paulis  # noqa: PLC0415

        paulis = []
        for port in io_ports:
            basis_set = {b.value for b in self.bases_at(port)}
            result = "I"
            for basis in basis_set:
                result = multiply_paulis(result, basis)
            paulis.append(result)

        return "".join(paulis)

    def external_stabilizer_on_graph(self, graph: BlockGraph) -> str:
        """Get the external stabilizer of the correlation surface on the graph.

        If the provided graph is an open graph, the external stabilizer is the Pauli
        operator supported on the input/output ports of the graph. Otherwise, the
        external stabilizer is the Pauli operator supported on the leaf nodes of the
        graph.

        Args:
            graph: The block graph to consider.

        Returns:
            The Pauli operator that is the external stabilizer of the correlation surface.

        """
        supports: list[Position3D]
        if graph.is_open:
            port_labels = graph.ordered_ports
            supports = [graph.ports[p] for p in port_labels]
        else:
            supports = [cube.position for cube in graph.leaf_cubes]
        zx = graph.to_zx_graph()
        p2v = zx.p2v
        zx_ports = [p2v[p] for p in supports]
        return self.external_stabilizer(zx_ports)

    def area(self) -> int:
        """Return the area of the correlation surface.

        The area of the correlation surface is the number of nodes it spans. A X node and a Z node
        with the same id are counted as two nodes.

        """
        span_nodes = {node for edge in self.span for node in edge}
        return len(span_nodes)

    def __xor__(self, other: CorrelationSurface) -> CorrelationSurface:
        return CorrelationSurface(self.span.symmetric_difference(other.span))



from __future__ import annotations

from collections.abc import Iterable
from functools import cached_property
from typing import Final, TypeGuard

from typing_extensions import override

from tqec.circuit.schedule.circuit import ScheduledCircuit
from tqec.compile.blocks.enums import SpatialBlockBorder
from tqec.compile.blocks.layers.atomic.base import BaseLayer
from tqec.compile.blocks.layers.atomic.plaquettes import PlaquetteLayer
from tqec.compile.blocks.positioning import (
    LayoutCubePosition2D,
    LayoutPipePosition2D,
    LayoutPosition2D,
)
from tqec.compile.generation import generate_circuit
from tqec.plaquette.plaquette import Plaquettes
from tqec.templates.enums import TemplateBorder
from tqec.templates.layout import LayoutTemplate
from tqec.utils.exceptions import TQECError
from tqec.utils.position import BlockPosition2D, Direction3D, Shift2D
from tqec.utils.scale import LinearFunction, PhysicalQubitScalable2D

DEFAULT_SHARED_QUBIT_DEPTH_AT_BORDER: Final[int] = 1
"""Default number of qubits that are shared between two neighbouring layers."""


def contains_only_plaquette_layers(
    layers: dict[LayoutPosition2D, BaseLayer],
) -> TypeGuard[dict[LayoutPosition2D, PlaquetteLayer]]:
    """Ensure correct typing when used in a conditional block."""
    return all(isinstance(layer, PlaquetteLayer) for layer in layers.values())


class LayoutLayer(BaseLayer):
    def __init__(
        self,
        layers: dict[LayoutPosition2D, BaseLayer],
        element_shape: PhysicalQubitScalable2D,
    ) -> None:
        """Glue several other layers together on a 2-dimensional grid.

        Args:
            layers: a mapping from positions on the 2-dimensional space to
                blocks implementing the circuit that should be present at that
                position.
                The mapping is expected to represent a connected computation.
            element_shape: scalable shape (in qubit coordinates) of each entry
                in the provided ``layers``.

        Raises:
            TQECError: if ``layers`` is empty.
            TQECError: if ``trimmed_spatial_borders`` is not empty.

        """
        super().__init__(frozenset())
        self._layers = layers
        self._element_shape = element_shape
        self._post_init_check()

    def _post_init_check(self) -> None:
        if not self.layers:
            raise TQECError(f"An instance of {type(self).__name__} should have at least one layer.")
        if self.trimmed_spatial_borders:
            raise TQECError(f"{LayoutLayer.__name__} cannot have trimmed spatial borders.")

    @property
    def layers(self) -> dict[LayoutPosition2D, BaseLayer]:
        """Return the layers composing ``self``."""
        return self._layers

    @property
    def element_shape(self) -> PhysicalQubitScalable2D:
        """Return the scalable shape of each stored elements."""
        return self._element_shape

    @cached_property
    def bounds(self) -> tuple[BlockPosition2D, BlockPosition2D]:
        """Get the top-left and bottom-right corners of the bounding box of ``self``.

        Returns:
            a tuple containing the corners of ``self``'s bounding box as positions
            containing the two minimum (top-left corner) or maximum (bottom-right
            corner) coordinates found in ``self``.

        """
        xs = [pos._x for pos in self.layers.keys()]
        ys = [pos._y for pos in self.layers.keys()]
        minx, maxx = min(xs), max(xs)
        miny, maxy = min(ys), max(ys)
        # We know for sure that the minima and maxima above are located on cube
        # positions and so are multiples of 2.
        return (
            BlockPosition2D(minx // 2, miny // 2),
            BlockPosition2D(maxx // 2, maxy // 2),
        )

    @property
    @override
    def scalable_shape(self) -> PhysicalQubitScalable2D:
        minp, maxp = self.bounds
        shapex, shapey = (maxp.x - minp.x) + 1, (maxp.y - minp.y) + 1
        return PhysicalQubitScalable2D(
            shapex * (self.element_shape.x - DEFAULT_SHARED_QUBIT_DEPTH_AT_BORDER)
            + DEFAULT_SHARED_QUBIT_DEPTH_AT_BORDER,
            shapey * (self.element_shape.y - DEFAULT_SHARED_QUBIT_DEPTH_AT_BORDER)
            + DEFAULT_SHARED_QUBIT_DEPTH_AT_BORDER,
        )

    @override
    def with_spatial_borders_trimmed(self, borders: Iterable[SpatialBlockBorder]) -> LayoutLayer:
        raise TQECError(f"Cannot trim spatial borders of a {type(self).__name__} instance.")

    def __eq__(self, value: object) -> bool:
        return (
            isinstance(value, LayoutLayer)
            and self.element_shape == value.element_shape
            and self.layers == value.layers
        )

    def __hash__(self) -> int:
        raise NotImplementedError(f"Cannot hash efficiently a {type(self).__name__}.")

    def to_template_and_plaquettes(self) -> tuple[LayoutTemplate, Plaquettes]:
        """Return an equivalent representation of ``self`` with a template and some plaquettes.

        Raises:
            NotImplementedError: if not all layers composing ``self`` are instances
                of :class:`~tqec.compile.blocks.layers.atomic.plaquette.PlaquetteLayer`.

        Returns:
            a tuple ``(template, plaquettes)`` that is ready to be used with
            :meth:`~tqec.compile.generation.generate_circuit` to obtain the quantum
            circuit representing ``self``.

        """
        if not contains_only_plaquette_layers(self.layers):
            raise NotImplementedError(
                f"Found a layer that is not an instance of {PlaquetteLayer.__name__}. "
                "Detector computation is not implemented (yet) for this case."
            )
        cubes: dict[BlockPosition2D, PlaquetteLayer] = {
            pos.to_block_position(): layer
            for pos, layer in self.layers.items()
            if isinstance(pos, LayoutCubePosition2D) and isinstance(layer, PlaquetteLayer)
        }
        template_dict: Final = {pos: layer.template for pos, layer in cubes.items()}
        plaquettes_dict = {pos: layer.plaquettes for pos, layer in cubes.items()}

        # Add plaquettes from each pipe to the plaquette_dict.
        pipes: dict[tuple[BlockPosition2D, BlockPosition2D], PlaquetteLayer] = {
            pos.to_pipe(): layer
            for pos, layer in self.layers.items()
            if isinstance(pos, LayoutPipePosition2D) and isinstance(layer, PlaquetteLayer)
        }
        for (u, v), pipe_layer in pipes.items():
            pipe_direction = Direction3D.from_neighbouring_positions(u.to_3d(), v.to_3d())
            # {u,v}_border: border of the respective node that is touched by the
            # the pipe.
            u_border: TemplateBorder
            v_border: TemplateBorder
            match pipe_direction:
                case Direction3D.X:
                    u_border, v_border = TemplateBorder.RIGHT, TemplateBorder.LEFT
                case Direction3D.Y:
                    u_border, v_border = TemplateBorder.BOTTOM, TemplateBorder.TOP
                case Direction3D.Z:
                    raise TQECError("Should not happen. This is a logical error.")

            # Updating plaquettes in plaquettes_dict
            for pos, (cube_border, pipe_border) in [
                (u, (u_border, v_border)),
                (v, (v_border, u_border)),
            ]:
                plaquette_indices_mapping = pipe_layer.template.get_border_indices(pipe_border).to(
                    template_dict[pos].get_border_indices(cube_border)
                )
                pipe_plaquette_collection = pipe_layer.plaquettes.collection
                plaquettes_dict[pos] = plaquettes_dict[pos].with_updated_plaquettes(
                    {
                        plaquette_indices_mapping[pipe_plaquette_index]: plaquette
                        for pipe_plaquette_index, plaquette in pipe_plaquette_collection.items()
                        # Filtering only plaquette indices that are on the side we are
                        # interested in.
                        if pipe_plaquette_index in plaquette_indices_mapping
                    }
                )

        template = LayoutTemplate(template_dict)
        return template, template.get_global_plaquettes(plaquettes_dict)

    def to_circuit(self, k: int) -> ScheduledCircuit:
        """Return the quantum circuit representing the layer.

        Args:
            k: scaling factor.

        Returns:
            quantum circuit representing the layer.

        """
        template, plaquettes = self.to_template_and_plaquettes()
        scheduled_circuit = generate_circuit(template, k, plaquettes)
        # Shift the qubits of the returned scheduled circuit
        mincube, _ = self.bounds
        eshape = self.element_shape.to_shape_2d(k)
        # See: https://github.com/tqec/tqec/issues/525
        # This is a temporary fix to the above issue, we may need a utility function
        # to calculate shift to avoid similar issues in the future.
        shift = Shift2D(mincube.x * (eshape.x - 1), mincube.y * (eshape.y - 1))
        shifted_circuit = scheduled_circuit.map_to_qubits(lambda q: q + shift)
        return shifted_circuit

    @property
    @override
    def scalable_num_moments(self) -> LinearFunction:
        return LinearFunction.unambiguous_max_on_positives(
            layer.scalable_num_moments for layer in self.layers.values()
        )

    @property
    def qubit_bounds(self) -> tuple[PhysicalQubitScalable2D, PhysicalQubitScalable2D]:
        """Return the top-left and bottom-right qubits representing the bounding box of ``self``.

        Returns:
            the ``(top_left, bottom_right)`` qubits.

        """
        tlb, brb = self.bounds
        eshape = self.element_shape
        increments = PhysicalQubitScalable2D(eshape.x, eshape.y) - (1, 1)
        tlq = PhysicalQubitScalable2D(tlb.x * increments.x, tlb.y * increments.y)
        brq = PhysicalQubitScalable2D((brb.x + 1) * increments.x, (brb.y + 1) * increments.y)
        # Note: for the moment, plaquette origin is defined as the CENTER of the
        # plaquette, which lead the above computations to be off-by-1. Correct that
        # before returning.
        shift = (-1, -1)
        return tlq + shift, brq + shift
